package iotruntime.master;

import iotruntime.*;
import iotruntime.slave.IoTAddress;
import iotruntime.slave.IoTDeviceAddress;
import iotruntime.messages.*;

// ASM packages
import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.ClassVisitor;

// Java packages
import java.io.*;
import java.util.*;
import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.IOException;
import java.lang.ClassNotFoundException;
import java.lang.Class;
import java.lang.reflect.*;
import java.net.Socket;
import java.net.ServerSocket;
import java.util.*;
import static java.lang.Math.toIntExact;

/** Class IoTMaster is responsible to use ClassRuntimeInstrumenterMaster
 *  to instrument the controller/device bytecode and starts multiple
 *  IoTSlave running on different JVM's in a distributed fashion.
 *
 * @author      Rahmadi Trimananda <rahmadi.trimananda @ uci.edu>
 * @version     1.0
 * @since       2016-06-16
 */
public class IoTMaster {

	/**
	 * IoTMaster class properties
	 * <p>
	 * CommunicationHandler maintains the data structure for hostnames and ports
	 * LoadBalancer assigns a job onto a host based on certain metrics
	 */
	private CommunicationHandler commHan;
	private LoadBalancer lbIoT;
	private RouterConfig routerConfig;
	private ObjectInitHandler objInitHand;
	private ObjectAddressInitHandler objAddInitHand;
	private String[] strObjectNames;
	private Map<String,ClassRuntimeInstrumenterMaster> mapClassNameToCrim;
	/**
	 * These properties hold information of a certain object
	 * at a certain time
	 */
	private String strObjName;
	private String strObjClassName;
	private String strObjClassInterfaceName;
	private String strObjStubClsIntfaceName;
	private String strIoTMasterHostAdd;
	private String strIoTSlaveControllerHostAdd;
	private String strIoTSlaveObjectHostAdd;
	private Class[] arrFieldClasses;
	private Object[] arrFieldValues;
	private Socket filesocket;

	// Constants that are to be extracted from config file
	private static String STR_MASTER_MAC_ADD;
	private static String STR_IOT_CODE_PATH;
	private static String STR_CONT_PATH;
	private static String STR_RUNTIME_DIR;
	private static String STR_CLS_PATH;
	private static String STR_RMI_PATH;
	private static String STR_RMI_HOSTNAME;
	private static String STR_LOG_FILE_PATH;
	private static String STR_SSH_USERNAME;
	private static String STR_ROUTER_ADD;
	private static String STR_MONITORING_HOST;
	private static String STR_ZB_GATEWAY_ADDRESS;
	private static String STR_ZB_GATEWAY_PORT;
	private static String STR_ZB_IOTMASTER_PORT;
	private static boolean BOOL_VERBOSE;

	/**
	 * IoTMaster class constants
	 * <p>
	 * Name constants - not to be configured by users
	 */
	private static final String STR_IOT_MASTER_NAME = "IoTMaster";
	private static final String STR_CFG_FILE_EXT = ".config";
	private static final String STR_CLS_FILE_EXT = ".class";
	private static final String STR_JAR_FILE_EXT = ".jar";
	private static final String STR_ZIP_FILE_EXT = ".zip";
	private static final String STR_TCP_PROTOCOL = "tcp";
	private static final String STR_UDP_PROTOCOL = "udp";
	private static final String STR_TCPGW_PROTOCOL = "tcpgw";
	private static final String STR_NO_PROTOCOL = "nopro";
	private static final String STR_SELF_MAC_ADD = "00:00:00:00:00:00";
	private static final String STR_INTERFACE_CLS_CFG = "INTERFACE_CLASS";
	private static final String STR_INT_STUB_CLS_CFG = "INTERFACE_STUB_CLASS";
	private static final String STR_FILE_TRF_CFG = "ADDITIONAL_ZIP_FILE";
	private static final String STR_YES = "Yes";
	private static final String STR_NO = "No";

	/**
	 * Runtime class name constants - not to be configured by users
	 */
	private static final String STR_REL_INSTRUMENTER_CLS = "iotruntime.master.RelationInstrumenter";
	private static final String STR_SET_INSTRUMENTER_CLS = "iotruntime.master.SetInstrumenter";
	private static final String STR_IOT_SLAVE_CLS = "iotruntime.slave.IoTSlave";
	private static final String STR_IOT_DEV_ADD_CLS = "IoTDeviceAddress";
	private static final String STR_IOT_ZB_ADD_CLS = "IoTZigbeeAddress";
	private static final String STR_IOT_ADD_CLS = "IoTAddress";
	
	/**
	 * Class constructor
	 *
	 */
	public IoTMaster(String[] argObjNms) {

		commHan = null;
		lbIoT = null;
		routerConfig = null;
		objInitHand = null;
		objAddInitHand = null;
		strObjectNames = argObjNms;
		strObjName = null;
		strObjClassName = null;
		strObjClassInterfaceName = null;
		strObjStubClsIntfaceName = null;
		strIoTMasterHostAdd = null;
		strIoTSlaveControllerHostAdd = null;
		strIoTSlaveObjectHostAdd = null;
		arrFieldClasses = null;
		arrFieldValues = null;
		filesocket = null;
		mapClassNameToCrim = null;

		STR_MASTER_MAC_ADD = null;
		STR_IOT_CODE_PATH = null;
		STR_CONT_PATH = null;
		STR_RUNTIME_DIR = null;
		STR_CLS_PATH = null;
		STR_RMI_PATH = null;
		STR_RMI_HOSTNAME = null;
		STR_LOG_FILE_PATH = null;
		STR_SSH_USERNAME = null;
		STR_ROUTER_ADD = null;
		STR_MONITORING_HOST = null;
		STR_ZB_GATEWAY_ADDRESS = null;
		STR_ZB_GATEWAY_PORT = null;
		STR_ZB_IOTMASTER_PORT = null;
		BOOL_VERBOSE = false;
	}

	/**
	 * A method to initialize CommunicationHandler, LoadBalancer, RouterConfig and ObjectInitHandler
	 *
	 * @return void
	 */
	private void initLiveDataStructure() {

		commHan = new CommunicationHandler(BOOL_VERBOSE);
		lbIoT = new LoadBalancer(BOOL_VERBOSE);
		lbIoT.setupLoadBalancer();
		routerConfig = new RouterConfig();
		routerConfig.getAddressList(STR_ROUTER_ADD);
		objInitHand = new ObjectInitHandler(BOOL_VERBOSE);
		objAddInitHand = new ObjectAddressInitHandler(BOOL_VERBOSE);
		mapClassNameToCrim = new HashMap<String,ClassRuntimeInstrumenterMaster>();
	}

	/**
	 * A method to initialize constants from config file
	 *
	 * @return void
	 */
	private void parseIoTMasterConfigFile() {
		// Parse configuration file
		Properties prop = new Properties();
		String strCfgFileName = STR_IOT_MASTER_NAME + STR_CFG_FILE_EXT;
		File file = new File(strCfgFileName);
		FileInputStream fis = null;
		try {
			fis = new FileInputStream(file);
			prop.load(fis);
			fis.close();
		} catch (IOException ex) {
			System.out.println("IoTMaster: Error reading config file: " + strCfgFileName);
			ex.printStackTrace();
		}
		// Initialize constants from config file
		STR_MASTER_MAC_ADD = prop.getProperty("MAC_ADDRESS");
		STR_IOT_CODE_PATH = prop.getProperty("IOT_CODE_PATH");
		STR_CONT_PATH = prop.getProperty("CONTROLLERS_CODE_PATH");
		STR_RUNTIME_DIR = prop.getProperty("RUNTIME_DIR");
		STR_CLS_PATH = prop.getProperty("CLASS_PATH");
		STR_RMI_PATH = prop.getProperty("RMI_PATH");
		STR_RMI_HOSTNAME = prop.getProperty("RMI_HOSTNAME");
		STR_LOG_FILE_PATH = prop.getProperty("LOG_FILE_PATH");
		STR_SSH_USERNAME = prop.getProperty("SSH_USERNAME");
		STR_ROUTER_ADD = prop.getProperty("ROUTER_ADD");
		STR_MONITORING_HOST = prop.getProperty("MONITORING_HOST");
		STR_ZB_GATEWAY_ADDRESS = prop.getProperty("ZIGBEE_GATEWAY_ADDRESS");
		STR_ZB_GATEWAY_PORT = prop.getProperty("ZIGBEE_GATEWAY_PORT");
		STR_ZB_IOTMASTER_PORT = prop.getProperty("ZIGBEE_IOTMASTER_PORT");
		if(prop.getProperty("VERBOSE").equals(STR_YES)) {
			BOOL_VERBOSE = true;
		}

		RuntimeOutput.print("IoTMaster: Extracting information from config file: " + strCfgFileName, BOOL_VERBOSE);
		RuntimeOutput.print("STR_MASTER_MAC_ADD=" + STR_MASTER_MAC_ADD, BOOL_VERBOSE);
		RuntimeOutput.print("STR_IOT_CODE_PATH=" + STR_IOT_CODE_PATH, BOOL_VERBOSE);
		RuntimeOutput.print("STR_CONT_PATH=" + STR_CONT_PATH, BOOL_VERBOSE);
		RuntimeOutput.print("STR_RUNTIME_DIR=" + STR_RUNTIME_DIR, BOOL_VERBOSE);
		RuntimeOutput.print("STR_CLS_PATH=" + STR_CLS_PATH, BOOL_VERBOSE);
		RuntimeOutput.print("STR_RMI_PATH=" + STR_RMI_PATH, BOOL_VERBOSE);
		RuntimeOutput.print("STR_RMI_HOSTNAME=" + STR_RMI_HOSTNAME, BOOL_VERBOSE);
		RuntimeOutput.print("STR_LOG_FILE_PATH=" + STR_LOG_FILE_PATH, BOOL_VERBOSE);
		RuntimeOutput.print("STR_SSH_USERNAME=" + STR_SSH_USERNAME, BOOL_VERBOSE);
		RuntimeOutput.print("STR_ROUTER_ADD=" + STR_ROUTER_ADD, BOOL_VERBOSE);
		RuntimeOutput.print("STR_MONITORING_HOST=" + STR_MONITORING_HOST, BOOL_VERBOSE);
		RuntimeOutput.print("STR_ZB_GATEWAY_ADDRESS=" + STR_ZB_GATEWAY_ADDRESS, BOOL_VERBOSE);
		RuntimeOutput.print("STR_ZB_GATEWAY_PORT=" + STR_ZB_GATEWAY_PORT, BOOL_VERBOSE);
		RuntimeOutput.print("STR_ZB_IOTMASTER_PORT=" + STR_ZB_IOTMASTER_PORT, BOOL_VERBOSE);
		RuntimeOutput.print("BOOL_VERBOSE=" + BOOL_VERBOSE, BOOL_VERBOSE);
		RuntimeOutput.print("IoTMaster: Information extracted successfully!", BOOL_VERBOSE);
	}

	/**
	 * A method to parse information from a config file
	 *
	 * @param	strCfgFileName	Config file name
	 * @param	strCfgField		Config file field name
	 * @return	String
	 */
	private String parseConfigFile(String strCfgFileName, String strCfgField) {
		// Parse configuration file
		Properties prop = new Properties();
		File file = new File(strCfgFileName);
		FileInputStream fis = null;
		try {
			fis = new FileInputStream(file);
			prop.load(fis);
			fis.close();
		} catch (IOException ex) {
			System.out.println("IoTMaster: Error reading config file: " + strCfgFileName);
			ex.printStackTrace();
		}
		System.out.println("IoTMaster: Reading " + strCfgField +
			" from config file: " + strCfgFileName + " with value: " + 
			prop.getProperty(strCfgField, null));
		// NULL is returned if the property isn't found
		return prop.getProperty(strCfgField, null);
	}

	/**
	 * A method to send files from IoTMaster
	 *
	 * @param  filesocket File socket object
	 * @param  sFileName  File name
	 * @param  lFLength   File length
	 * @return            void
	 */
	private void sendFile(Socket filesocket, String sFileName, long lFLength) throws IOException {

		File file = new File(sFileName);
		byte[] bytFile = new byte[toIntExact(lFLength)];
		InputStream inFileStream = new FileInputStream(file);

		OutputStream outFileStream = filesocket.getOutputStream();
		int iCount;
		while ((iCount = inFileStream.read(bytFile)) > 0) {
			outFileStream.write(bytFile, 0, iCount);
		}
		filesocket.close();
		RuntimeOutput.print("IoTMaster: File sent!", BOOL_VERBOSE);
	}

	/**
	 * A method to create a thread
	 *
	 * @param  sSSHCmd    SSH command
	 * @return            void
	 */
	private void createThread(String sSSHCmd) throws IOException {

		// Start a new thread to start a new JVM
		new Thread() {
			Runtime runtime = Runtime.getRuntime();
			Process process = runtime.exec(sSSHCmd);
		}.start();
		RuntimeOutput.print("Executing: " + sSSHCmd, BOOL_VERBOSE);
	}

	/**
	 * A method to send command from master and receive reply from slave
	 *
	 * @params  msgSend     Message object
	 * @params  strPurpose  String that prints purpose message
	 * @params  inStream    Input stream
	 * @params  outStream   Output stream
	 * @return  void
	 */
	private void commMasterToSlave(Message msgSend, String strPurpose,
		ObjectInputStream inStream, ObjectOutputStream outStream) 
			throws IOException, ClassNotFoundException {

		// Send message/command from master
		outStream.writeObject(msgSend);
		RuntimeOutput.print("IoTMaster: Send message: " + strPurpose, BOOL_VERBOSE);

		// Get reply from slave as acknowledgment
		Message msgReply = (Message) inStream.readObject();
		RuntimeOutput.print("IoTMaster: Reply message: " + msgReply.getMessage(), BOOL_VERBOSE);
	}

	/**
	 * A private method to instrument IoTSet device
	 *
	 * @params  strFieldIdentifier        String field name + object ID
	 * @params  strFieldName              String field name
	 * @params  strIoTSlaveObjectHostAdd  String slave host address
	 * @params  inStream                  ObjectInputStream communication
	 * @params  inStream                  ObjectOutputStream communication
	 * @return  void
	 */
	private void instrumentIoTSetDevice(String strFieldIdentifier, String strFieldName, String strIoTSlaveObjectHostAdd,
		ObjectInputStream inStream, ObjectOutputStream outStream) 
			throws IOException, ClassNotFoundException, InterruptedException {

		// Get information from the set
		List<Object[]> listObject = objAddInitHand.getFields(strFieldIdentifier);
		// Create a new IoTSet
		Message msgCrtIoTSet = new MessageCreateSetRelation(IoTCommCode.CREATE_NEW_IOTSET, strFieldName);
		commMasterToSlave(msgCrtIoTSet, "Create new IoTSet for IoTDeviceAddress!", inStream, outStream);
		int iRows = listObject.size();
		RuntimeOutput.print("IoTMaster: Number of rows for IoTDeviceAddress: " + iRows, BOOL_VERBOSE);
		// Transfer the address
		for(int iRow=0; iRow<iRows; iRow++) {
			arrFieldValues = listObject.get(iRow);
			// Get device address - if 00:00:00:00:00:00 that means it needs the driver object address (self)
			String strDeviceAddress = null;
			if (arrFieldValues[0].equals(STR_SELF_MAC_ADD)) {
				strDeviceAddress = strIoTSlaveObjectHostAdd;
			} else {
				strDeviceAddress = routerConfig.getIPFromMACAddress((String) arrFieldValues[0]);
			}
			int iDestDeviceDriverPort = (int) arrFieldValues[1];
			String strProtocol = (String) arrFieldValues[2];
			// Check for wildcard feature			
			boolean bSrcPortWildCard = false;
			boolean bDstPortWildCard = false;
			if (arrFieldValues.length > 3) {
				bSrcPortWildCard = (boolean) arrFieldValues[3];
				bDstPortWildCard = (boolean) arrFieldValues[4];
			}
			// Add the port connection into communication handler - if it's not assigned yet
			if (commHan.getComPort(strDeviceAddress) == null) {
				commHan.addPortConnection(strIoTSlaveObjectHostAdd, strDeviceAddress);
			}
			// Send address one by one
			Message msgGetIoTSetObj = new MessageGetDeviceObject(IoTCommCode.GET_DEVICE_IOTSET_OBJECT,
				strDeviceAddress, commHan.getComPort(strDeviceAddress), iDestDeviceDriverPort,
				bSrcPortWildCard, bDstPortWildCard);
			commMasterToSlave(msgGetIoTSetObj, "Get IoTSet objects!", inStream, outStream);
		}
		// Reinitialize IoTSet on device object
		commMasterToSlave(new MessageSimple(IoTCommCode.REINITIALIZE_IOTSET_FIELD),
			"Reinitialize IoTSet fields!", inStream, outStream);
	}


	/**
	 * A private method to instrument IoTSet Zigbee device
	 *
	 * @params  Map.Entry<String,Object>  Entry of map IoTSet instrumentation
	 * @params  strFieldName              String field name
	 * @params  strIoTSlaveObjectHostAdd  String slave host address
	 * @params  inStream                  ObjectInputStream communication
	 * @params  inStream                  ObjectOutputStream communication
	 * @return  void
	 */
	private void instrumentIoTSetZBDevice(Map.Entry<String,Object> map, String strFieldName, String strIoTSlaveObjectHostAdd,
		ObjectInputStream inStream, ObjectOutputStream outStream) 
			throws IOException, ClassNotFoundException, InterruptedException {

		// Get information from the set
		SetInstrumenter setInstrumenter = (SetInstrumenter) map.getValue();
		// Create a new IoTSet
		Message msgCrtIoTSet = new MessageCreateSetRelation(IoTCommCode.CREATE_NEW_IOTSET, strFieldName);
		commMasterToSlave(msgCrtIoTSet, "Create new IoTSet for IoTZigbeeAddress!", inStream, outStream);
		// Prepare ZigbeeConfig
		String strZigbeeGWAddress = routerConfig.getIPFromMACAddress(STR_ZB_GATEWAY_ADDRESS);
		int iZigbeeGWPort = Integer.parseInt(STR_ZB_GATEWAY_PORT);
		int iZigbeeIoTMasterPort = Integer.parseInt(STR_ZB_IOTMASTER_PORT);
		commHan.addDevicePort(iZigbeeIoTMasterPort);
		ZigbeeConfig zbConfig = new ZigbeeConfig(strZigbeeGWAddress, iZigbeeGWPort, iZigbeeIoTMasterPort, 
			BOOL_VERBOSE);
		// Add the port connection into communication handler - if it's not assigned yet
		if (commHan.getComPort(strZigbeeGWAddress) == null) {
			commHan.addPortConnection(strIoTSlaveObjectHostAdd, strZigbeeGWAddress);
		}		
		int iRows = setInstrumenter.numberOfRows();
		RuntimeOutput.print("IoTMaster: Number of rows for IoTZigbeeAddress: " + iRows, BOOL_VERBOSE);
		// Transfer the address
		for(int iRow=0; iRow<iRows; iRow++) {
			arrFieldValues = setInstrumenter.fieldValues(iRow);
			// Get device address
			String strZBDevAddress = (String) arrFieldValues[0];
			// Send policy to Zigbee gateway - TODO: Need to clear policy first?
			zbConfig.setPolicy(strIoTSlaveObjectHostAdd, commHan.getComPort(strZigbeeGWAddress), strZBDevAddress);
			// Send address one by one
			Message msgGetIoTSetZBObj = new MessageGetSimpleDeviceObject(IoTCommCode.GET_ZB_DEV_IOTSET_OBJECT,
				strZBDevAddress);
			commMasterToSlave(msgGetIoTSetZBObj, "Get IoTSet objects!", inStream, outStream);
		}
		zbConfig.closeConnection();
		// Reinitialize IoTSet on device object
		commMasterToSlave(new MessageSimple(IoTCommCode.REINITIALIZE_IOTSET_FIELD),
											"Reinitialize IoTSet fields!", inStream, outStream);
	}

	
	/**
	 * A private method to instrument IoTSet of addresses
	 *
	 * @params  strFieldIdentifier        String field name + object ID
	 * @params  strFieldName              String field name
	 * @params  inStream                  ObjectInputStream communication
	 * @params  inStream                  ObjectOutputStream communication
	 * @return  void
	 */
	private void instrumentIoTSetAddress(String strFieldIdentifier, String strFieldName,
		ObjectInputStream inStream, ObjectOutputStream outStream) 
			throws IOException, ClassNotFoundException, InterruptedException {

		// Get information from the set
		List<Object[]> listObject = objAddInitHand.getFields(strFieldIdentifier);
		// Create a new IoTSet
		Message msgCrtIoTSet = new MessageCreateSetRelation(IoTCommCode.CREATE_NEW_IOTSET, strFieldName);
		commMasterToSlave(msgCrtIoTSet, "Create new IoTSet for IoTAddress!", inStream, outStream);
		int iRows = listObject.size();
		RuntimeOutput.print("IoTMaster: Number of rows for IoTAddress: " + iRows, BOOL_VERBOSE);
		// Transfer the address
		for(int iRow=0; iRow<iRows; iRow++) {
			arrFieldValues = listObject.get(iRow);
			// Get device address
			String strAddress = (String) arrFieldValues[0];
			// Send address one by one
			Message msgGetIoTSetAddObj = new MessageGetSimpleDeviceObject(IoTCommCode.GET_ADD_IOTSET_OBJECT,
				strAddress);
			commMasterToSlave(msgGetIoTSetAddObj, "Get IoTSet objects!", inStream, outStream);
		}
		// Reinitialize IoTSet on device object
		commMasterToSlave(new MessageSimple(IoTCommCode.REINITIALIZE_IOTSET_FIELD),
											"Reinitialize IoTSet fields!", inStream, outStream);
	}


	/**
	 * A private method to instrument an object on a specific machine and setting up policies
	 *
	 * @params  strFieldObjectID  String field object ID
	 * @return  void
	 */
	private void instrumentObject(String strFieldObjectID) throws IOException {

		// Extract the interface name for RMI
		// e.g. ProximitySensorInterface, TempSensorInterface, etc.
		
String strObjCfgFile = STR_IOT_CODE_PATH + strObjClassName + "/" + strObjClassName + STR_CFG_FILE_EXT;
		strObjClassInterfaceName = parseConfigFile(strObjCfgFile, STR_INTERFACE_CLS_CFG);
		strObjStubClsIntfaceName = parseConfigFile(strObjCfgFile, STR_INT_STUB_CLS_CFG);
		// Create an object name, e.g. ProximitySensorImplPS1
		strObjName = strObjClassName + strFieldObjectID;
		// Check first if host exists
		if(commHan.objectExists(strObjName)) {
			// If this object exists already ...
			// Re-read IoTSlave object hostname for further reference
			strIoTSlaveObjectHostAdd = commHan.getHostAddress(strObjName);
			RuntimeOutput.print("IoTMaster: Object with name: " + strObjName + " has existed!", BOOL_VERBOSE);
		} else {
			// If this is a new object ... then create one
			// Get host address for IoTSlave from LoadBalancer
			//strIoTSlaveObjectHostAdd = lbIoT.selectHost();
			strIoTSlaveObjectHostAdd = routerConfig.getIPFromMACAddress(lbIoT.selectHost());
			if (strIoTSlaveControllerHostAdd == null)
				throw new Error("IoTMaster: Could not translate MAC to IP address! Please check the router's /tmp/dhcp.leases!");
			RuntimeOutput.print("IoTMaster: Object name: " + strObjName, BOOL_VERBOSE);
			// Add port connection and get port numbers
			// Naming for objects ProximitySensor becomes ProximitySensor0, ProximitySensor1, etc.
			commHan.addPortConnection(strIoTSlaveObjectHostAdd, strObjName);
			commHan.addActiveControllerObject(strFieldObjectID, strObjName, strObjClassName, strObjClassInterfaceName, 
				strObjStubClsIntfaceName, strIoTSlaveObjectHostAdd, arrFieldValues, arrFieldClasses);
			// ROUTING POLICY: IoTMaster and device/controller object
			// Master-slave communication
			routerConfig.configureRouterMainPolicies(STR_ROUTER_ADD, strIoTMasterHostAdd,
				strIoTSlaveObjectHostAdd, STR_TCP_PROTOCOL, commHan.getComPort(strObjName));
			// ROUTING POLICY: Send the same routing policy to both the hosts
			routerConfig.configureHostMainPolicies(strIoTMasterHostAdd, strIoTMasterHostAdd,
				strIoTSlaveObjectHostAdd, STR_TCP_PROTOCOL, commHan.getComPort(strObjName));
			routerConfig.configureHostMainPolicies(strIoTSlaveObjectHostAdd, strIoTMasterHostAdd,
				strIoTSlaveObjectHostAdd, STR_TCP_PROTOCOL, commHan.getComPort(strObjName));
			// Need to accommodate callback functions here - open ports for TCP
			routerConfig.configureRouterMainPolicies(STR_ROUTER_ADD, strIoTSlaveControllerHostAdd,
				strIoTSlaveObjectHostAdd, STR_TCP_PROTOCOL);
			routerConfig.configureHostMainPolicies(strIoTSlaveControllerHostAdd, strIoTSlaveControllerHostAdd,
				strIoTSlaveObjectHostAdd, STR_TCP_PROTOCOL);
			routerConfig.configureHostMainPolicies(strIoTSlaveObjectHostAdd, strIoTSlaveControllerHostAdd,
				strIoTSlaveObjectHostAdd, STR_TCP_PROTOCOL);
			// Instrument the IoTSet declarations inside the class file
			instrumentObjectIoTSet(strFieldObjectID);
		}
		// Send routing policy to router for controller object
		// ROUTING POLICY: RMI communication - RMI registry and stub ports
		routerConfig.configureRouterMainPolicies(STR_ROUTER_ADD, strIoTSlaveControllerHostAdd, strIoTSlaveObjectHostAdd,
			STR_TCP_PROTOCOL, commHan.getRMIRegPort(strObjName));
		routerConfig.configureRouterMainPolicies(STR_ROUTER_ADD, strIoTSlaveControllerHostAdd, strIoTSlaveObjectHostAdd,
			STR_TCP_PROTOCOL, commHan.getRMIStubPort(strObjName));
		// Send the same set of routing policies to compute nodes
		routerConfig.configureHostMainPolicies(strIoTSlaveControllerHostAdd, strIoTSlaveControllerHostAdd, strIoTSlaveObjectHostAdd,
			STR_TCP_PROTOCOL, commHan.getRMIRegPort(strObjName));
		routerConfig.configureHostMainPolicies(strIoTSlaveObjectHostAdd, strIoTSlaveControllerHostAdd, strIoTSlaveObjectHostAdd,
			STR_TCP_PROTOCOL, commHan.getRMIRegPort(strObjName));
		routerConfig.configureHostMainPolicies(strIoTSlaveControllerHostAdd, strIoTSlaveControllerHostAdd, strIoTSlaveObjectHostAdd,
			STR_TCP_PROTOCOL, commHan.getRMIStubPort(strObjName));
		routerConfig.configureHostMainPolicies(strIoTSlaveObjectHostAdd, strIoTSlaveControllerHostAdd, strIoTSlaveObjectHostAdd,
			STR_TCP_PROTOCOL, commHan.getRMIStubPort(strObjName));
	}

	/**
	 * A private method to set router policies for IoTDeviceAddress objects
	 *
	 * @params  strFieldIdentifier        String field name + object ID
	 * @params  Map.Entry<String,Object>  Entry of map IoTSet instrumentation
	 * @params  strIoTSlaveObjectHostAdd  String slave host address
	 * @return  void
	 */
	private void setRouterPolicyIoTSetDevice(String strFieldIdentifier, Map.Entry<String,Object> map, 
		String strIoTSlaveObjectHostAdd) {

		// Get information from the set
		SetInstrumenter setInstrumenter = (SetInstrumenter) map.getValue();
		int iRows = setInstrumenter.numberOfRows();
		RuntimeOutput.print("IoTMaster: Number of rows for IoTDeviceAddress: " + iRows, BOOL_VERBOSE);
		// Transfer the address
		for(int iRow=0; iRow<iRows; iRow++) {
			arrFieldValues = setInstrumenter.fieldValues(iRow);
			objAddInitHand.addField(strFieldIdentifier, arrFieldValues);	// Save this for object instantiation
			// Get device address - if 00:00:00:00:00:00 that means it needs the driver object address (self)
			String strDeviceAddress = null;
			if (arrFieldValues[0].equals(STR_SELF_MAC_ADD)) {
				strDeviceAddress = strIoTSlaveObjectHostAdd;
			} else {
				strDeviceAddress = routerConfig.getIPFromMACAddress((String) arrFieldValues[0]);
			}
			int iDestDeviceDriverPort = (int) arrFieldValues[1];
			String strProtocol = (String) arrFieldValues[2];
			// Add the port connection into communication handler - if it's not assigned yet
			if (commHan.getComPort(strDeviceAddress) == null) {
				commHan.addPortConnection(strIoTSlaveObjectHostAdd, strDeviceAddress);
			}
			// Send routing policy to router for device drivers and devices
			// ROUTING POLICY: RMI communication - RMI registry and stub ports
			if((iDestDeviceDriverPort == -1) && (!strProtocol.equals(STR_NO_PROTOCOL))) {
				// Port number -1 means that we don't set the policy strictly to port number level
				// "nopro" = no protocol specified for just TCP or just UDP (can be both used as well)
				// ROUTING POLICY: Device driver and device
				routerConfig.configureRouterMainPolicies(STR_ROUTER_ADD, strIoTSlaveObjectHostAdd, strDeviceAddress,
					strProtocol);
				// ROUTING POLICY: Send to the compute node where the device driver is
				routerConfig.configureHostMainPolicies(strIoTSlaveObjectHostAdd, strIoTSlaveObjectHostAdd,
					strDeviceAddress, strProtocol);
			} else if((iDestDeviceDriverPort == -1) && (strProtocol.equals(STR_NO_PROTOCOL))) {
				routerConfig.configureRouterMainPolicies(STR_ROUTER_ADD, strIoTSlaveObjectHostAdd, strDeviceAddress);
				routerConfig.configureHostMainPolicies(strIoTSlaveObjectHostAdd, strIoTSlaveObjectHostAdd, strDeviceAddress);
			} else if(strProtocol.equals(STR_TCPGW_PROTOCOL)) {
				// This is a TCP protocol that connects, e.g. a phone to our runtime system
				// that provides a gateway access (accessed through destination port number)
				commHan.addDevicePort(iDestDeviceDriverPort);
				routerConfig.configureRouterMainPolicies(STR_ROUTER_ADD, strIoTSlaveObjectHostAdd, strDeviceAddress,
					STR_TCP_PROTOCOL, iDestDeviceDriverPort);
				routerConfig.configureHostMainPolicies(strIoTSlaveObjectHostAdd, strIoTSlaveObjectHostAdd, strDeviceAddress,
					STR_TCP_PROTOCOL, iDestDeviceDriverPort);
				routerConfig.configureRouterHTTPPolicies(STR_ROUTER_ADD, strIoTSlaveObjectHostAdd, strDeviceAddress);
				routerConfig.configureHostHTTPPolicies(strIoTSlaveObjectHostAdd, strIoTSlaveObjectHostAdd, strDeviceAddress);
			} else {
				// Other port numbers...
				commHan.addDevicePort(iDestDeviceDriverPort);
				routerConfig.configureRouterMainPolicies(STR_ROUTER_ADD, strIoTSlaveObjectHostAdd, strDeviceAddress,
					strProtocol, commHan.getComPort(strDeviceAddress), iDestDeviceDriverPort);
				routerConfig.configureHostMainPolicies(strIoTSlaveObjectHostAdd, strIoTSlaveObjectHostAdd, strDeviceAddress,
					strProtocol, commHan.getComPort(strDeviceAddress), iDestDeviceDriverPort);
			}
		}
	}

	/**
	 * A private method to set router policies for IoTAddress objects
	 *
	 * @params  strFieldIdentifier        String field name + object ID
	 * @params  Map.Entry<String,Object>  Entry of map IoTSet instrumentation
	 * @params  strHostAddress            String host address
	 * @return  void
	 */
	private void setRouterPolicyIoTSetAddress(String strFieldIdentifier, Map.Entry<String,Object> map, 
		String strHostAddress) {

		// Get information from the set
		SetInstrumenter setInstrumenter = (SetInstrumenter) map.getValue();
		int iRows = setInstrumenter.numberOfRows();
		RuntimeOutput.print("IoTMaster: Number of rows for IoTAddress: " + iRows, BOOL_VERBOSE);
		// Transfer the address
		for(int iRow=0; iRow<iRows; iRow++) {
			arrFieldValues = setInstrumenter.fieldValues(iRow);
			objAddInitHand.addField(strFieldIdentifier, arrFieldValues);	// Save this for object instantiation
			// Get device address
			String strAddress = (String) arrFieldValues[0];
			// Setting up router policies for HTTP/HTTPs
			routerConfig.configureRouterHTTPPolicies(STR_ROUTER_ADD, strHostAddress, strAddress);
			routerConfig.configureHostHTTPPolicies(strHostAddress, strHostAddress, strAddress);
		}
	}

	/**
	 * A private method to instrument an object's IoTSet and IoTRelation field to up policies
	 * <p>
	 * Mostly the IoTSet fields would contain IoTDeviceAddress objects
	 *
	 * @params  strFieldObjectID  String field object ID
	 * @return  void
	 */
	private void instrumentObjectIoTSet(String strFieldObjectID) throws IOException {

		// If this is a new object ... then create one
		// Instrument the class source code and look for IoTSet for device addresses
		// e.g. @config private IoTSet<IoTDeviceAddress> lb_addresses;
		String strObjectClassNamePath = STR_IOT_CODE_PATH + strObjClassName + "/" + strObjClassName + STR_CLS_FILE_EXT;
		FileInputStream fis = new FileInputStream(strObjectClassNamePath);
		ClassReader cr = new ClassReader(fis);
		ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_FRAMES);
		// We need Object ID to instrument IoTDeviceAddress
		ClassRuntimeInstrumenterMaster crim = new ClassRuntimeInstrumenterMaster(cw, strFieldObjectID, BOOL_VERBOSE);
		cr.accept(crim, 0);
		fis.close();
		RuntimeOutput.print("IoTMaster: Going to instrument for " + strObjClassName + " with objectID " + 
			strFieldObjectID, BOOL_VERBOSE);
		// Get the object and the class names
		// Build objects for IoTSet and IoTRelation fields in the device object classes
		mapClassNameToCrim.put(strObjClassName + strFieldObjectID, crim);
		HashMap<String,Object> hmObjectFieldObjects = crim.getFieldObjects();
		for(Map.Entry<String,Object> map : hmObjectFieldObjects.entrySet()) {
			RuntimeOutput.print("IoTMaster: Object name: " + map.getValue().getClass().getName(), BOOL_VERBOSE);
			// Iterate over HashMap and choose between processing
			String strFieldName = map.getKey();
			String strClassName = map.getValue().getClass().getName();
			String strFieldIdentifier = strFieldName + strFieldObjectID;
			if(strClassName.equals(STR_SET_INSTRUMENTER_CLS)) {
				SetInstrumenter setInstrumenter = (SetInstrumenter) map.getValue();
				if(setInstrumenter.getObjTableName().equals(STR_IOT_DEV_ADD_CLS)) { 
				// Instrument the normal IoTDeviceAddress
					setRouterPolicyIoTSetDevice(strFieldIdentifier, map, strIoTSlaveObjectHostAdd);
				} else if(setInstrumenter.getObjTableName().equals(STR_IOT_ADD_CLS)) { 
				// Instrument the IoTAddress
					setRouterPolicyIoTSetAddress(strFieldIdentifier, map, strIoTSlaveObjectHostAdd);
				} else if(setInstrumenter.getObjTableName().equals(STR_IOT_ZB_ADD_CLS)) { 
				// Instrument the IoTZigbeeAddress - special feature for Zigbee device support
					RuntimeOutput.print("IoTMaster: IoTZigbeeAddress found! No router policy is set here..", 
						BOOL_VERBOSE);
				} else {
					String strErrMsg = "IoTMaster: Device driver object" +
										" can only have IoTSet<IoTAddress>, IoTSet<IoTDeviceAddress>," +
										" or IoTSet<IoTZigbeeAddress>!";
					throw new Error(strErrMsg);
				}
			} else {
				String strErrMsg = "IoTMaster: Device driver object can only have IoTSet for addresses!";
				throw new Error(strErrMsg);
			}
		}
	}


	/**
	 * A private method to create an object on a specific machine
	 *
	 * @params  strObjName  				String object name
	 * @params  strObjClassName 			String object class name
	 * @params  strObjClassInterfaceName 	String object class interface name
	 * @params  strIoTSlaveObjectHostAdd 	String IoTSlave host address
	 * @params  strFieldObjectID  			String field object ID
	 * @params  arrFieldValues				Array of field values
	 * @params  arrFieldClasses				Array of field classes
	 * @return  void
	 */
	private void createObject(String strObjName, String strObjClassName, String strObjClassInterfaceName, String strObjStubClsIntfaceName,
		String strIoTSlaveObjectHostAdd, String strFieldObjectID, Object[] arrFieldValues, Class[] arrFieldClasses) 
		throws IOException, FileNotFoundException, ClassNotFoundException, InterruptedException {

		// PROFILING
		long start = 0;
		long result = 0;

		// PROFILING
		start = System.currentTimeMillis();

		// Construct ssh command line
		// e.g. ssh rtrimana@dw-2.eecs.uci.edu cd <path>;
		//      java -cp $CLASSPATH:./*.jar
		//           -Djava.rmi.server.codebase=file:./*.jar
		//           iotruntime.IoTSlave dw-1.eecs.uci.edu 46151 23829 42874 &
		// The In-Port for IoTMaster is the Out-Port for IoTSlave and vice versa
		String strSSHCommand = STR_SSH_USERNAME + strIoTSlaveObjectHostAdd + " cd " + STR_RUNTIME_DIR + " sudo java " +
			STR_CLS_PATH + " " + STR_RMI_PATH + " " + STR_RMI_HOSTNAME +
			strIoTSlaveObjectHostAdd + " " + STR_IOT_SLAVE_CLS + " " + strIoTMasterHostAdd + " " +
			commHan.getComPort(strObjName) + " " + commHan.getRMIRegPort(strObjName) + " " +
			commHan.getRMIStubPort(strObjName) + " >& " + STR_LOG_FILE_PATH + strObjName + ".log &";
		RuntimeOutput.print(strSSHCommand, BOOL_VERBOSE);
		// Start a new thread to start a new JVM
		createThread(strSSHCommand);
		ServerSocket serverSocket = new ServerSocket(commHan.getComPort(strObjName));
		Socket socket = serverSocket.accept();
		ObjectInputStream inStream = new ObjectInputStream(socket.getInputStream());
		ObjectOutputStream outStream = new ObjectOutputStream(socket.getOutputStream());

		// PROFILING
		result = System.currentTimeMillis()-start;
		System.out.println("\n\n ==> Time needed to start JVM for " + strObjName + ": " + result + "\n\n");

		// PROFILING
		start = System.currentTimeMillis();

		// Create message to transfer file first
		String sFileName = strObjClassName + STR_JAR_FILE_EXT;
		String sPath = STR_IOT_CODE_PATH + strObjClassName + "/" + sFileName;
		File file = new File(sPath);
		commMasterToSlave(new MessageSendFile(IoTCommCode.TRANSFER_FILE, sFileName, file.length()),
			"Sending file!", inStream, outStream);
		// Send file - JAR file for object creation
		sendFile(serverSocket.accept(), sPath, file.length());
		Message msgReply = (Message) inStream.readObject();
		RuntimeOutput.print("IoTMaster: Reply message: " + msgReply.getMessage(), BOOL_VERBOSE);

		// PROFILING
		result = System.currentTimeMillis()-start;
		System.out.println("\n\n ==> Time needed to send JAR file for " + strObjName + ": " + result + "\n\n");

		// PROFILING
		start = System.currentTimeMillis();

		// Pack object information to create object on a IoTSlave
		Message msgObjIoTSlave = new MessageCreateObject(IoTCommCode.CREATE_OBJECT, strIoTSlaveObjectHostAdd,
			strObjClassName, strObjName, strObjClassInterfaceName, strObjStubClsIntfaceName, commHan.getRMIRegPort(strObjName), 
			commHan.getRMIStubPort(strObjName), arrFieldValues, arrFieldClasses);
		// Send message
		commMasterToSlave(msgObjIoTSlave, "Sending object information", inStream, outStream);
		// Instrument the class source code and look for IoTSet for device addresses
		// e.g. @config private IoTSet<IoTDeviceAddress> lb_addresses;
		RuntimeOutput.print("IoTMaster: Instantiating for " + strObjClassName + " with objectID " + 
			strFieldObjectID, BOOL_VERBOSE);
		// Get the object and the class names
		// Build objects for IoTSet and IoTRelation fields in the device object classes
		ClassRuntimeInstrumenterMaster crim = mapClassNameToCrim.get(strObjClassName + strFieldObjectID);
		HashMap<String,Object> hmObjectFieldObjects = crim.getFieldObjects();
		for(Map.Entry<String,Object> map : hmObjectFieldObjects.entrySet()) {
			RuntimeOutput.print("IoTMaster: Object name: " + map.getValue().getClass().getName(), BOOL_VERBOSE);
			// Iterate over HashMap and choose between processing
			String strFieldName = map.getKey();
			String strClassName = map.getValue().getClass().getName();
			String strFieldIdentifier = strFieldName + strFieldObjectID;
			if(strClassName.equals(STR_SET_INSTRUMENTER_CLS)) {
				SetInstrumenter setInstrumenter = (SetInstrumenter) map.getValue();
				if(setInstrumenter.getObjTableName().equals(STR_IOT_DEV_ADD_CLS)) { 
				// Instrument the normal IoTDeviceAddress
					synchronized(this) {
						instrumentIoTSetDevice(strFieldIdentifier, strFieldName, strIoTSlaveObjectHostAdd, inStream, outStream);
					}
				} else if(setInstrumenter.getObjTableName().equals(STR_IOT_ZB_ADD_CLS)) { 
				// Instrument the IoTZigbeeAddress - special feature for Zigbee device support
					synchronized(this) {
						instrumentIoTSetZBDevice(map, strFieldName, strIoTSlaveObjectHostAdd, inStream, outStream);
					}
				} else if(setInstrumenter.getObjTableName().equals(STR_IOT_ADD_CLS)) { 
				// Instrument the IoTAddress
					synchronized(this) {
						instrumentIoTSetAddress(strFieldIdentifier, strFieldName, inStream, outStream);
					}
				} else {
					String strErrMsg = "IoTMaster: Device driver object" +
										" can only have IoTSet<IoTAddress>, IoTSet<IoTDeviceAddress>," +
										" or IoTSet<IoTZigbeeAddress>!";
					throw new Error(strErrMsg);
				}
			} else {
				String strErrMsg = "IoTMaster: Device driver object can only have IoTSet for addresses!";
				throw new Error(strErrMsg);
			}
		}
		// End the session
		// TODO: Change this later
		outStream.writeObject(new MessageSimple(IoTCommCode.END_SESSION));

		// PROFILING
		result = System.currentTimeMillis()-start;
		System.out.println("\n\n ==> Time needed to create object " + strObjName + " and instrument IoTDeviceAddress: " + result + "\n\n");

		// Closing streams
		outStream.close();
		inStream.close();
		socket.close();
		serverSocket.close();
	}


	/**
	 * A private method to create controller objects
	 *
	 * @return  void
	 */
	private void createControllerObjects() throws InterruptedException {

		// Create a list of threads
		List<Thread> threads = new ArrayList<Thread>();
		// Get the list of active controller objects and loop it
		List<String> listActiveControllerObject = commHan.getActiveControllerObjectList();
		for(String strObjName : listActiveControllerObject) {

			ObjectCreationInfo objCrtInfo = commHan.getObjectCreationInfo(strObjName);
			Thread objectThread = new Thread(new Runnable() {
				public void run() {
					synchronized(this) {
						try {
							createObject(strObjName, objCrtInfo.getObjectClassName(), objCrtInfo.getObjectClassInterfaceName(),
								objCrtInfo.getObjectStubClassInterfaceName(), objCrtInfo.getIoTSlaveObjectHostAdd(), 
								commHan.getFieldObjectID(strObjName), commHan.getArrayFieldValues(strObjName), 
								commHan.getArrayFieldClasses(strObjName));
						} catch (IOException 			| 
								 ClassNotFoundException |
								 InterruptedException ex) {
							ex.printStackTrace();
						}
					}
				}
			});
			threads.add(objectThread);
			objectThread.start();
		}
		// Join all threads
		for (Thread thread : threads) {
			try {
				thread.join();
			} catch (InterruptedException ex) {
				ex.printStackTrace();
			}
		}
	}	


	/**
	 * A private method to instrument IoTSet
	 *
	 * @params  Map.Entry<String,Object>  Entry of map IoTSet instrumentation
	 * @params  strFieldName              String field name
	 * @return  void
	 */
	private void instrumentIoTSet(Map.Entry<String,Object> map, String strFieldName) 
		throws IOException, ClassNotFoundException, InterruptedException {
				
		// Get information from the set
		SetInstrumenter setInstrumenter = (SetInstrumenter) map.getValue();
		objInitHand.addField(strFieldName, IoTCommCode.CREATE_NEW_IOTSET);

		int iRows = setInstrumenter.numberOfRows();
		for(int iRow=0; iRow<iRows; iRow++) {
			// Get field classes and values
			arrFieldClasses = setInstrumenter.fieldClasses(iRow);
			arrFieldValues = setInstrumenter.fieldValues(iRow);
			// Get object ID and class name
			String strObjID = setInstrumenter.fieldObjectID(iRow);
			strObjClassName = setInstrumenter.fieldEntryType(strObjID);
			// Call the method to create an object
			instrumentObject(strObjID);
			objInitHand.addObjectIntoField(strFieldName, strIoTSlaveObjectHostAdd, strObjName,
				strObjClassName, strObjClassInterfaceName, strObjStubClsIntfaceName, commHan.getRMIRegPort(strObjName), 
				commHan.getRMIStubPort(strObjName));
		}
	}


	/**
	 * A private method to instrument IoTRelation
	 *
	 * @params  Map.Entry<String,Object>  Entry of map IoTRelation instrumentation
	 * @params  strFieldName              String field name
	 * @return  void
	 */
	private void instrumentIoTRelation(Map.Entry<String,Object> map, String strFieldName) 
		throws IOException, ClassNotFoundException, InterruptedException {

			// Get information from the set
		RelationInstrumenter relationInstrumenter = (RelationInstrumenter) map.getValue();
		int iRows = relationInstrumenter.numberOfRows();
		objInitHand.addField(strFieldName, IoTCommCode.CREATE_NEW_IOTRELATION);

		for(int iRow=0; iRow<iRows; iRow++) {
			// Operate on the first set first
			arrFieldClasses = relationInstrumenter.firstFieldClasses(iRow);
			arrFieldValues = relationInstrumenter.firstFieldValues(iRow);
			String strObjID = relationInstrumenter.firstFieldObjectID(iRow);
			strObjClassName = relationInstrumenter.firstEntryFieldType(strObjID);
			// Call the method to create an object
			instrumentObject(strObjID);
			// Get the first object controller host address
			String strFirstIoTSlaveObjectHostAdd = strIoTSlaveObjectHostAdd;
			objInitHand.addObjectIntoField(strFieldName, strIoTSlaveObjectHostAdd, strObjName,
				strObjClassName, strObjClassInterfaceName, strObjStubClsIntfaceName, 
				commHan.getRMIRegPort(strObjName), commHan.getRMIStubPort(strObjName));
			// Operate on the second set
			arrFieldClasses = relationInstrumenter.secondFieldClasses(iRow);
			arrFieldValues = relationInstrumenter.secondFieldValues(iRow);
			strObjID = relationInstrumenter.secondFieldObjectID(iRow);
			strObjClassName = relationInstrumenter.secondEntryFieldType(strObjID);
			// Call the method to create an object
			instrumentObject(strObjID);
			// Get the second object controller host address
			String strSecondIoTSlaveObjectHostAdd = strIoTSlaveObjectHostAdd;
			objInitHand.addSecondObjectIntoField(strFieldName, strIoTSlaveObjectHostAdd, strObjName,
				strObjClassName, strObjClassInterfaceName, strObjStubClsIntfaceName, 
				commHan.getRMIRegPort(strObjName), commHan.getRMIStubPort(strObjName));

			// ROUTING POLICY: first and second controller objects in IoTRelation
			routerConfig.configureRouterMainPolicies(STR_ROUTER_ADD, strFirstIoTSlaveObjectHostAdd,
				strSecondIoTSlaveObjectHostAdd, STR_TCP_PROTOCOL);
			// ROUTING POLICY: Send the same routing policy to both the hosts
			routerConfig.configureHostMainPolicies(strFirstIoTSlaveObjectHostAdd, strFirstIoTSlaveObjectHostAdd,
				strSecondIoTSlaveObjectHostAdd, STR_TCP_PROTOCOL);
			routerConfig.configureHostMainPolicies(strSecondIoTSlaveObjectHostAdd, strFirstIoTSlaveObjectHostAdd,
				strSecondIoTSlaveObjectHostAdd, STR_TCP_PROTOCOL);
		}
	}

	/**
	 * A method to reinitialize IoTSet and IoTRelation in the code based on ObjectInitHandler information
	 *
	 * @params  inStream                  ObjectInputStream communication
	 * @params  outStream                 ObjectOutputStream communication
	 * @return	void
	 */
	private void initializeSetsAndRelations(ObjectInputStream inStream, ObjectOutputStream outStream) 
		throws IOException, ClassNotFoundException {
		// Get list of fields
		List<String> strFields = objInitHand.getListOfFields();
		// Iterate on HostAddress
		for(String str : strFields) {
			IoTCommCode iotcommMsg = objInitHand.getFieldMessage(str);
			if (iotcommMsg == IoTCommCode.CREATE_NEW_IOTSET) {
				// == COMMUNICATION WITH IOTSLAVE CONTROLLER TO CREATE IOTSET
				Message msgCrtIoTSet = new MessageCreateSetRelation(IoTCommCode.CREATE_NEW_IOTSET, str);
				commMasterToSlave(msgCrtIoTSet, "Create new IoTSet!", inStream, outStream);
				List<ObjectInitInfo> listObject = objInitHand.getListObjectInitInfo(str);
				for (ObjectInitInfo objInitInfo : listObject) {
					// == COMMUNICATION WITH IOTSLAVE CONTROLLER TO FILL IN IOTSET
					commMasterToSlave(new MessageGetObject(IoTCommCode.GET_IOTSET_OBJECT, objInitInfo.getIoTSlaveObjectHostAdd(),
						objInitInfo.getObjectName(), objInitInfo.getObjectClassName(), objInitInfo.getObjectClassInterfaceName(), 
						objInitInfo.getObjectStubClassInterfaceName(), objInitInfo.getRMIRegistryPort(), objInitInfo.getRMIStubPort()), 
						"Get IoTSet object!", inStream, outStream);	
				}
				// == COMMUNICATION WITH IOTSLAVE CONTROLLER TO REINITIALIZE IOTSET FIELD
				commMasterToSlave(new MessageSimple(IoTCommCode.REINITIALIZE_IOTSET_FIELD),
					"Renitialize IoTSet field!", inStream, outStream);
			} else if (iotcommMsg == IoTCommCode.CREATE_NEW_IOTRELATION) {
				// == COMMUNICATION WITH IOTSLAVE CONTROLLER TO CREATE IOTRELATION
				Message msgCrtIoTRel = new MessageCreateSetRelation(IoTCommCode.CREATE_NEW_IOTRELATION, str);
				commMasterToSlave(msgCrtIoTRel, "Create new IoTRelation!", inStream, outStream);
				List<ObjectInitInfo> listObject = objInitHand.getListObjectInitInfo(str);
				List<ObjectInitInfo> listSecondObject = objInitHand.getSecondObjectInitInfo(str);
				Iterator it = listSecondObject.iterator();
				for (ObjectInitInfo objInitInfo : listObject) {
					// == COMMUNICATION WITH IOTSLAVE CONTROLLER TO FILL IN IOTRELATION (FIRST OBJECT)
					commMasterToSlave(new MessageGetObject(IoTCommCode.GET_IOTRELATION_FIRST_OBJECT, 
						objInitInfo.getIoTSlaveObjectHostAdd(), objInitInfo.getObjectName(), objInitInfo.getObjectClassName(),
						objInitInfo.getObjectClassInterfaceName(), objInitInfo.getObjectStubClassInterfaceName(),
						objInitInfo.getRMIRegistryPort(), objInitInfo.getRMIStubPort()), 
						"Get IoTRelation first object!", inStream, outStream);
					ObjectInitInfo objSecObj = (ObjectInitInfo) it.next();
					// == COMMUNICATION WITH IOTSLAVE CONTROLLER TO FILL IN IOTRELATION (SECOND OBJECT)
					commMasterToSlave(new MessageGetObject(IoTCommCode.GET_IOTRELATION_SECOND_OBJECT,
						objSecObj.getIoTSlaveObjectHostAdd(), objSecObj.getObjectName(), objSecObj.getObjectClassName(),
						objSecObj.getObjectClassInterfaceName(), objInitInfo.getObjectStubClassInterfaceName(),
						objSecObj.getRMIRegistryPort(), objSecObj.getRMIStubPort()), 
						"Get IoTRelation second object!", inStream, outStream);
				}
				// == COMMUNICATION WITH IOTSLAVE CONTROLLER TO REINITIALIZE IOTRELATION FIELD
				commMasterToSlave(new MessageSimple(IoTCommCode.REINITIALIZE_IOTRELATION_FIELD),
					"Renitialize IoTRelation field!", inStream, outStream);
			}
		}
	}

	/**
	 * A method to set router basic policies at once
	 *
	 * @param	strRouter String router name
	 * @return	void
	 */
	private void setRouterBasicPolicies(String strRouter) {

		String strMonitorHost = routerConfig.getIPFromMACAddress(STR_MONITORING_HOST);
		routerConfig.configureRouterICMPPolicies(strRouter, strMonitorHost);
		routerConfig.configureRouterDHCPPolicies(strRouter);
		routerConfig.configureRouterDNSPolicies(strRouter);
		routerConfig.configureRouterSSHPolicies(strRouter, strMonitorHost);
		routerConfig.configureRejectPolicies(strRouter);
	}

	/**
	 * A method to set host basic policies at once
	 *
	 * @param	strHost String host name
	 * @return	void
	 */
	private void setHostBasicPolicies(String strHost) {

		String strMonitorHost = routerConfig.getIPFromMACAddress(STR_MONITORING_HOST);
		routerConfig.configureHostDHCPPolicies(strHost);
		routerConfig.configureHostDNSPolicies(strHost);
		if (strHost.equals(strMonitorHost)) {
		// Check if this is the monitoring host
			routerConfig.configureHostICMPPolicies(strHost);
			routerConfig.configureHostSSHPolicies(strHost);
		} else {
			routerConfig.configureHostICMPPolicies(strHost, strMonitorHost);
			routerConfig.configureHostSSHPolicies(strHost, strMonitorHost);
		}
		// Apply SQL allowance policies to master host
		if (strHost.equals(strIoTMasterHostAdd)) {
			routerConfig.configureHostSQLPolicies(strHost);
		}
		routerConfig.configureRejectPolicies(strHost);
	}

	/**
	 * A method to create a thread for policy deployment
	 *
	 * @param  strRouterAddress		String router address to configure
	 * @param  setHostAddresses		Set of strings for host addresses to configure
	 * @return            			void
	 */
	private void createPolicyThreads(String strRouterAddress, Set<String> setHostAddresses) throws IOException {

		// Create a list of threads
		List<Thread> threads = new ArrayList<Thread>();
		// Start threads for hosts
		for(String strAddress : setHostAddresses) {
			Thread policyThread = new Thread(new Runnable() {
				public void run() {
					synchronized(this) {
						routerConfig.sendHostPolicies(strAddress);
					}
				}
			});
			threads.add(policyThread);
			policyThread.start();
			RuntimeOutput.print("Deploying policies for: " + strAddress, BOOL_VERBOSE);
		}
		// A thread for router
		Thread policyThread = new Thread(new Runnable() {
			public void run() {
				synchronized(this) {
					routerConfig.sendRouterPolicies(strRouterAddress);
				}
			}
		});
		threads.add(policyThread);
		policyThread.start();
		RuntimeOutput.print("Deploying policies on router: " + strRouterAddress, BOOL_VERBOSE);		
		// Join all threads
		for (Thread thread : threads) {
			try {
				thread.join();
			} catch (InterruptedException ex) {
				ex.printStackTrace();
			}
		}
	}


	/**
	 * A method to assign objects to multiple JVMs, including
	 * the controller/device object that uses other objects
	 * in IoTSet and IoTRelation
	 *
	 * @return       void
	 */
	private void createObjects() {

		// PROFILING
		long start = 0;
		long result = 0;

		try {
			// Extract hostname for this IoTMaster from MySQL DB
			strIoTMasterHostAdd = routerConfig.getIPFromMACAddress(STR_MASTER_MAC_ADD);
			// Loop as we can still find controller/device classes
			for(int i=0; i<strObjectNames.length; i++) {
				// PROFILING
				start = System.currentTimeMillis();

				// Assign a new list of PrintWriter objects
				routerConfig.renewPrintWriter();
				// Get controller names one by one
				String strObjControllerName = strObjectNames[i];
				// Use LoadBalancer to assign a host address
				//strIoTSlaveControllerHostAdd = lbIoT.selectHost();
				strIoTSlaveControllerHostAdd = routerConfig.getIPFromMACAddress(lbIoT.selectHost());
				if (strIoTSlaveControllerHostAdd == null)
					throw new Error("IoTMaster: Could not translate MAC to IP address! Please check the router's /tmp/dhcp.leases!");
				// == START INITIALIZING CONTROLLER/DEVICE IOTSLAVE ==
				// Add port connection and get port numbers
				// Naming for objects ProximitySensor becomes ProximitySensor0, ProximitySensor1, etc.
				commHan.addPortConnection(strIoTSlaveControllerHostAdd, strObjControllerName);
				// ROUTING POLICY: IoTMaster and main controller object
				routerConfig.configureRouterMainPolicies(STR_ROUTER_ADD, strIoTMasterHostAdd,
					strIoTSlaveControllerHostAdd, STR_TCP_PROTOCOL, commHan.getComPort(strObjControllerName));
				// ROUTING POLICY: Send the same routing policy to both the hosts
				routerConfig.configureHostMainPolicies(strIoTMasterHostAdd, strIoTMasterHostAdd,
					strIoTSlaveControllerHostAdd, STR_TCP_PROTOCOL, commHan.getComPort(strObjControllerName));
				routerConfig.configureHostMainPolicies(strIoTSlaveControllerHostAdd, strIoTMasterHostAdd,
					strIoTSlaveControllerHostAdd, STR_TCP_PROTOCOL, commHan.getComPort(strObjControllerName));

				// Construct ssh command line and create a controller thread for e.g. AcmeProximity
				String strSSHCommand = STR_SSH_USERNAME + strIoTSlaveControllerHostAdd + " cd " +
					STR_RUNTIME_DIR + " sudo java " + STR_CLS_PATH + " " +
					STR_RMI_PATH + " " + STR_IOT_SLAVE_CLS + " " + strIoTMasterHostAdd + " " +
					commHan.getComPort(strObjControllerName) + " " +
					commHan.getRMIRegPort(strObjControllerName) + " " +
					commHan.getRMIStubPort(strObjControllerName) + " >& " +
					STR_LOG_FILE_PATH + strObjControllerName + ".log &";
				RuntimeOutput.print(strSSHCommand, BOOL_VERBOSE);
				createThread(strSSHCommand);
				// Wait for connection
				// Create a new socket for communication
				ServerSocket serverSocket = new ServerSocket(commHan.getComPort(strObjControllerName));
				Socket socket = serverSocket.accept();
				ObjectInputStream inStream = new ObjectInputStream(socket.getInputStream());
				ObjectOutputStream outStream = new ObjectOutputStream(socket.getOutputStream());
				RuntimeOutput.print("IoTMaster: Communication established!", BOOL_VERBOSE);

				// PROFILING
				result = System.currentTimeMillis()-start;
				System.out.println("\n\n ==> From start until after SSH for main controller: " + result);
				// PROFILING
				start = System.currentTimeMillis();

				// Send files for every controller class
				// e.g. AcmeProximity.jar and AcmeProximity.zip
				String strControllerClassName = strObjControllerName + STR_CLS_FILE_EXT;
				String strControllerClassNamePath = STR_CONT_PATH + strObjControllerName + "/" +
					strControllerClassName;
				// Send .jar file
				String strControllerJarName = strObjControllerName + STR_JAR_FILE_EXT;
				String strControllerJarNamePath = STR_CONT_PATH + strObjControllerName + "/" +
					strControllerJarName;
				File file = new File(strControllerJarNamePath);
				commMasterToSlave(new MessageSendFile(IoTCommCode.TRANSFER_FILE, strControllerJarName, file.length()),
					"Sending file!", inStream, outStream);
				// Send file - Class file for object creation
				sendFile(serverSocket.accept(), strControllerJarNamePath, file.length());
				Message msgReply = (Message) inStream.readObject();
				RuntimeOutput.print("IoTMaster: Reply message: " + msgReply.getMessage(), BOOL_VERBOSE);
				// Send .zip file if additional zip file is specified
				String strObjCfgFile = strObjControllerName + STR_CFG_FILE_EXT;
				String strObjCfgFilePath = STR_CONT_PATH + strObjControllerName + "/" + strObjCfgFile;
				String strAdditionalFile = parseConfigFile(strObjCfgFilePath, STR_FILE_TRF_CFG);
				if (strAdditionalFile.equals(STR_YES)) {
					String strControllerCmpName = strObjControllerName + STR_ZIP_FILE_EXT;
					String strControllerCmpNamePath = STR_CONT_PATH + strObjControllerName + "/" +
						strControllerCmpName;
					file = new File(strControllerCmpNamePath);
					commMasterToSlave(new MessageSendFile(IoTCommCode.TRANSFER_FILE, strControllerCmpName, file.length()),
						"Sending file!", inStream, outStream);
					// Send file - Class file for object creation
					sendFile(serverSocket.accept(), strControllerCmpNamePath, file.length());
					msgReply = (Message) inStream.readObject();
					RuntimeOutput.print("IoTMaster: Reply message: " + msgReply.getMessage(), BOOL_VERBOSE);
				}
				// Create main controller/device object
				commMasterToSlave(new MessageCreateMainObject(IoTCommCode.CREATE_MAIN_OBJECT, strObjControllerName),
					"Create main object!", inStream, outStream);

				// PROFILING
				result = System.currentTimeMillis()-start;
				System.out.println("\n\n ==> From IoTSlave start until main controller object is created: " + result);
				System.out.println(" ==> Including file transfer times!\n\n");
				// PROFILING
				start = System.currentTimeMillis();

				// == END INITIALIZING CONTROLLER/DEVICE IOTSLAVE ==
				// Instrumenting one file
				RuntimeOutput.print("IoTMaster: Opening class file: " + strControllerClassName, BOOL_VERBOSE);
				RuntimeOutput.print("IoTMaster: Class file path: " + strControllerClassNamePath, BOOL_VERBOSE);
				FileInputStream fis = new FileInputStream(strControllerClassNamePath);
				ClassReader cr = new ClassReader(fis);
				ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_FRAMES);
				ClassRuntimeInstrumenterMaster crim = new ClassRuntimeInstrumenterMaster(cw, null, BOOL_VERBOSE);
				cr.accept(crim, 0);
				fis.close();
				// Get the object and the class names
				// Build objects for IoTSet and IoTRelation fields in the controller/device classes
				HashMap<String,Object> hmControllerFieldObjects = crim.getFieldObjects();
				for(Map.Entry<String,Object> map : hmControllerFieldObjects.entrySet()) {
					RuntimeOutput.print("IoTMaster: Object name: " + map.getValue().getClass().getName(), BOOL_VERBOSE);
					// Iterate over HashMap and choose between processing
					// SetInstrumenter vs. RelationInstrumenter
					String strFieldName = map.getKey();
					String strClassName = map.getValue().getClass().getName();
					if(strClassName.equals(STR_SET_INSTRUMENTER_CLS)) {
						SetInstrumenter setInstrumenter = (SetInstrumenter) map.getValue();
						if(setInstrumenter.getObjTableName().equals(STR_IOT_DEV_ADD_CLS)) { 
							String strErrMsg = "IoTMaster: Controller object" +
								" cannot have IoTSet<IoTDeviceAddress>!";
							throw new Error(strErrMsg);
						} else if(setInstrumenter.getObjTableName().equals(STR_IOT_ZB_ADD_CLS)) { 
							String strErrMsg = "IoTMaster: Controller object" +
								" cannot have IoTSet<ZigbeeAddress>!";
							throw new Error(strErrMsg);
						} else if(setInstrumenter.getObjTableName().equals(STR_IOT_ADD_CLS)) { 
						// Instrument the IoTAddress
							setRouterPolicyIoTSetAddress(strFieldName, map, strIoTSlaveControllerHostAdd);
							instrumentIoTSetAddress(strFieldName, strFieldName, inStream, outStream);
						} else {
						// Any other cases
							instrumentIoTSet(map, strFieldName);
						}
					} else if (strClassName.equals(STR_REL_INSTRUMENTER_CLS)) {
						instrumentIoTRelation(map, strFieldName);
					}
				}
				// PROFILING
				result = System.currentTimeMillis()-start;
				System.out.println("\n\n ==> Time needed to instrument device driver objects: " + result + "\n\n");
				System.out.println(" ==> #Objects: " + commHan.getActiveControllerObjectList().size() + "\n\n");

				// PROFILING
				start = System.currentTimeMillis();

				// ROUTING POLICY: Deploy basic policies if this is the last controller
				if (i == strObjectNames.length-1) {
					// ROUTING POLICY: implement basic policies to reject all other irrelevant traffics
					for(String s: commHan.getHosts()) {
						setHostBasicPolicies(s);
					}
					// We retain all the basic policies for router, 
					// but we delete the initial allowance policies for internal all TCP and UDP communications
					setRouterBasicPolicies(STR_ROUTER_ADD);
				}
				// Close access to policy files and deploy policies
				routerConfig.close();
				// Deploy the policy
				HashSet<String> setAddresses = new HashSet<String>(commHan.getHosts());
				setAddresses.add(strIoTMasterHostAdd);
				createPolicyThreads(STR_ROUTER_ADD, setAddresses);

				// PROFILING
				result = System.currentTimeMillis()-start;
				System.out.println("\n\n ==> Time needed to send policy files and deploy them : " + result + "\n\n");

				// PROFILING
				start = System.currentTimeMillis();

				// Separating object creations and Set/Relation initializations
				createControllerObjects();

				// PROFILING
				result = System.currentTimeMillis()-start;
				System.out.println("\n\n ==> Time needed to instantiate objects: " + result + "\n\n");
				// PROFILING
				start = System.currentTimeMillis();

				// Sets and relations initializations
				initializeSetsAndRelations(inStream, outStream);

				// PROFILING
				result = System.currentTimeMillis()-start;
				System.out.println("\n\n ==> Time needed to initialize sets and relations: " + result + "\n\n");

				// == COMMUNICATION WITH IOTSLAVE CONTROLLER TO EXECUTE INIT METHOD
				commMasterToSlave(new MessageSimple(IoTCommCode.INVOKE_INIT_METHOD),
					"Invoke init() method!", inStream, outStream);
				// == COMMUNICATION WITH IOTSLAVE CONTROLLER TO END PROCESS
				outStream.writeObject(new MessageSimple(IoTCommCode.END_SESSION));
				outStream.close();
				inStream.close();
				socket.close();
				serverSocket.close();
				commHan.printLists();
				lbIoT.printHostInfo();
			}

		} catch (IOException          |
				 InterruptedException |
				 ClassNotFoundException ex) {
			System.out.println("IoTMaster: Exception: "
				+ ex.getMessage());
			ex.printStackTrace();
		}
	}

	public static void main(String args[]) {

		// Detect the available controller/device classes
		// Input args[] should be used to list the controllers/devices
		// e.g. java IoTMaster AcmeProximity AcmeThermostat AcmeVentController
		IoTMaster iotMaster = new IoTMaster(args);
		// Read config file
		iotMaster.parseIoTMasterConfigFile();
		// Initialize CommunicationHandler, LoadBalancer, and RouterConfig
		iotMaster.initLiveDataStructure();
		// Create objects
		iotMaster.createObjects();
	}
}
